@@ -37,9 +37,12 @@ static void convert_hiragana_to_katakana (const WideString &hira,
 static int skk_key_mask = SCIM_KEY_ControlMask | SCIM_KEY_AltMask;
 
 
+
 SKKCore::SKKCore      (KeyBind *keybind, SKKAutomaton *key2kana,
-                       SKKDictionary *dict)
+                       SKKDictionary *dict, History &hist)
     : m_keybind(keybind),
+      m_hist(hist),
+      m_histmgr(hist),
       m_dict(dict),
       m_skk_mode(SKK_MODE_HIRAGANA),
       m_input_mode(INPUT_MODE_DIRECT),
@@ -201,6 +204,7 @@ SKKCore::commit_or_preedit (const WideString &str)
             m_preeditstr += str;
         }
         m_preedit_pos += str.length();
+        m_histmgr.clear();
         break;
     case INPUT_MODE_OKURI:
         m_okuristr += str;
@@ -209,7 +213,8 @@ SKKCore::commit_or_preedit (const WideString &str)
             m_dict->lookup(m_preeditstr+m_okurihead, true, m_ltable);
             if (m_ltable.empty()) {
                 set_input_mode(INPUT_MODE_LEARNING);
-                m_learning = new SKKCore(m_keybind, m_key2kana, m_dict);
+                m_learning = new SKKCore(m_keybind, m_key2kana,
+                                         m_dict, m_hist);
             } else {
                 set_input_mode(INPUT_MODE_CONVERTING);
             }
@@ -410,6 +415,7 @@ SKKCore::clear_preedit (void)
     m_preedit_pos = 0;
     m_okuristr.clear();
     m_okurihead = 0;
+    m_histmgr.clear();
 }
 
 void
@@ -458,6 +464,9 @@ SKKCore::action_kakutei (void)
             } else {
                 commit_string(m_preeditstr);
             }
+            if (m_input_mode == INPUT_MODE_PREEDIT) {
+                m_hist.add_entry(m_preeditstr);
+            }
             clear_preedit();
         }
         clear_pending();
@@ -522,15 +531,16 @@ SKKCore::action_convert (void)
         retval = action_nextpage();
         if (!retval) {
             set_input_mode(INPUT_MODE_LEARNING);
-            m_learning = new SKKCore(m_keybind, m_key2kana, m_dict);
+            m_learning = new SKKCore(m_keybind, m_key2kana, m_dict, m_hist);
         }
         return true;
     case INPUT_MODE_PREEDIT:
         clear_pending();
+        m_hist.add_entry(m_preeditstr);
         m_dict->lookup(m_preeditstr, false, m_ltable);
         if (m_ltable.empty()) {
             set_input_mode(INPUT_MODE_LEARNING);
-            m_learning = new SKKCore(m_keybind, m_key2kana, m_dict);
+            m_learning = new SKKCore(m_keybind, m_key2kana, m_dict, m_hist);
         } else {
             set_input_mode(INPUT_MODE_CONVERTING);
         }
@@ -572,6 +582,9 @@ SKKCore::action_katakana (bool half)
             } else {
                 commit_string(m_preeditstr);
             }
+            if (!m_preeditstr.empty() && m_input_mode == INPUT_MODE_PREEDIT) {
+                m_hist.add_entry(m_preeditstr);
+            }
             clear_preedit();
             clear_pending();
             set_input_mode(INPUT_MODE_DIRECT);
@@ -607,6 +620,9 @@ SKKCore::action_toggle_case (void)
                 *i = tolower(code);
             }
         }
+        if (!m_preeditstr.empty()) {
+            m_hist.add_entry(m_preeditstr);
+        }
         commit_string(m_preeditstr);
         clear_preedit();
         clear_pending();
@@ -643,6 +659,36 @@ SKKCore::action_start_preedit (void)
     }
 }
 
+bool
+SKKCore::action_completion (void)
+{
+    if (m_input_mode == INPUT_MODE_PREEDIT) {
+        if (m_histmgr.is_clear()) {
+            m_histmgr.setup_completion(m_preeditstr);
+        } else {
+            m_histmgr.next_cand();
+        }
+        m_histmgr.get_current_candidate(m_preeditstr);
+        m_preedit_pos = m_preeditstr.size();
+        return true;
+    } else {
+        return false;
+    }
+}
+
+bool
+SKKCore::action_completion_back (void)
+{
+    if (m_input_mode == INPUT_MODE_PREEDIT) {
+        if (m_histmgr.prev_cand()) {
+            m_histmgr.get_current_candidate(m_preeditstr);
+            m_preedit_pos = m_preeditstr.size();
+            return true;
+        }
+    } 
+    return false;
+}
+
 bool
 SKKCore::action_prevcand (void)
 {
@@ -713,6 +759,7 @@ SKKCore::action_backspace (void)
                 action_cancel();
             } else {
                 m_preeditstr.erase(m_preedit_pos-1, 1);
+                m_histmgr.clear();
                 m_preedit_pos--;
             }
             break;
@@ -753,8 +800,10 @@ SKKCore::action_delete (void)
             m_ltable.clear();
             break;
         case INPUT_MODE_PREEDIT:
-            if (m_preedit_pos < m_preeditstr.length())
+            if (m_preedit_pos < m_preeditstr.length()) {
                 m_preeditstr.erase(m_preedit_pos, 1);
+                m_histmgr.clear();
+            }
             break;
         case INPUT_MODE_DIRECT:
             if (m_commitstr.empty()) {
@@ -782,7 +831,8 @@ SKKCore::action_forward (void)
         if (m_ltable.visible_table()) {
             if (!m_ltable.cursor_down()) {
                 set_input_mode(INPUT_MODE_LEARNING);
-                m_learning = new SKKCore(m_keybind, m_key2kana, m_dict);
+                m_learning = new SKKCore(m_keybind, m_key2kana,
+                                         m_dict, m_hist);
             }
             return true;
         } else {
@@ -798,6 +848,7 @@ SKKCore::action_forward (void)
         }
     case INPUT_MODE_PREEDIT:
         clear_pending();
+        m_histmgr.clear();
         if (m_preedit_pos < m_preeditstr.length()) {
             m_preedit_pos++;
         } else if (m_commit_pos < m_commitstr.length()) {
@@ -828,6 +879,7 @@ SKKCore::action_backward (void)
         }
     case INPUT_MODE_DIRECT:
         clear_pending();
+        m_histmgr.clear();
         if (m_commit_pos > 0) {
             m_commit_pos--;
             return true;
@@ -866,6 +918,7 @@ SKKCore::action_home (void)
         }
     case INPUT_MODE_PREEDIT:
         clear_pending();
+        m_histmgr.clear();
         if (m_preedit_pos > 0) {
             m_preedit_pos = 0;
         } else if (m_commit_pos > 0) {
@@ -895,6 +948,7 @@ SKKCore::action_end (void)
         }
     case INPUT_MODE_PREEDIT:
         clear_pending();
+        m_histmgr.clear();
         if (m_preedit_pos < m_preeditstr.length()) {
             m_preedit_pos = m_preeditstr.length();
         } else if (m_commit_pos < m_commitstr.length()) {
@@ -991,6 +1045,12 @@ SKKCore::process_remaining_keybinds (const KeyEvent &key)
     if (m_keybind->match_end_keys(key))
         return action_end();
 
+    if (m_keybind->match_completion_keys(key))
+        return action_completion();
+
+    if (m_keybind->match_completion_back_keys(key))
+        return action_completion_back();
+
     return false;
 }
 
@@ -1177,13 +1237,18 @@ SKKCore::process_key_event (const KeyEvent key)
         }
         commit_converting();
         set_input_mode(INPUT_MODE_DIRECT);
+        if (key.code == SCIM_KEY_Return &&
+            (ignore_return ^ (key.mask & SCIM_KEY_ShiftMask != 0))) {
+            return true;
+        }
     }
 
     if (m_input_mode == INPUT_MODE_LEARNING) {
         bool retval = m_learning->process_key_event(key);
         char code = key.get_ascii_code();
         if (key.code == SCIM_KEY_Return || m_learning->m_end_flag) {
-            if (ignore_return && key.code == SCIM_KEY_Return)
+            if (key.code == SCIM_KEY_Return &&
+                (ignore_return ^ (key.mask & SCIM_KEY_ShiftMask != 0)))
                 retval = true;
             if (m_learning->m_commitstr.empty()) {
                 /* learning is canceled */
@@ -1244,6 +1309,17 @@ SKKCore::process_key_event (const KeyEvent key)
         return retval;
     }
 
+    if (m_input_mode == INPUT_MODE_PREEDIT && key.code == SCIM_KEY_Return &&
+        !(key.mask & skk_key_mask)) {
+        action_kakutei();
+        if (ignore_return ^ (key.mask & SCIM_KEY_ShiftMask != 0)) {
+            return true;
+        } else {
+            return false;
+        }
+    }
+
+
     switch (m_skk_mode) {
     case SKK_MODE_ASCII:
         return process_ascii(key);
