@@ -31,6 +31,8 @@
 #  include <config.h>
 #endif
 
+#include <vector>
+
 #include <scim.h>
 #include <gtk/scimkeyselection.h>
 #include "scim_skk_prefs.h"
@@ -150,6 +152,7 @@ struct ColorConfigData
 };
 
 // Internal data declaration.
+static std::vector<String> __config_sysdicts;
 static String __config_userdict     = SCIM_SKK_CONFIG_USERDICT_DEFAULT;
 static int    __config_listsize     = SCIM_SKK_CONFIG_CANDVEC_SIZE_DEFAULT;
 static bool   __config_annot_view   = SCIM_SKK_CONFIG_ANNOT_VIEW_DEFAULT;
@@ -162,8 +165,9 @@ static String __config_annot_pos    = SCIM_SKK_CONFIG_ANNOT_POS_DEFAULT;
 static String __config_annot_target = SCIM_SKK_CONFIG_ANNOT_TARGET_DEFAULT;
 static String __config_selection_style = SCIM_SKK_CONFIG_SELECTION_STYLE_DEFAULT;
 
-static bool __have_changed    = true;
+bool __have_changed    = true;
 
+GtkWidget           * __widget_sysdicts        = 0;
 static GtkWidget    * __widget_userdict        = 0;
 static GtkWidget    * __widget_listsize        = 0;
 static GtkWidget    * __widget_annot_view      = 0;
@@ -194,6 +198,46 @@ static KeyboardConfigData __config_keyboards_common [] =
         NULL,
         NULL,
     },
+    {
+        SCIM_SKK_CONFIG_UPCASE_KEY,
+        SCIM_SKK_CONFIG_UPCASE_KEY_DEFAULT,
+        N_("Upcase Keys:"),
+        N_("Select Upcase Keys"),
+        N_("The key events to toggle case of characters in preedit string. "),
+        NULL,
+        NULL,
+    },
+    {
+        SCIM_SKK_CONFIG_COMPLETION_KEY,
+        SCIM_SKK_CONFIG_COMPLETION_KEY_DEFAULT,
+        N_("Completion Keys:"),
+        N_("Select Completion Keys"),
+        N_("The key events to complete preedit string. "),
+        NULL,
+        NULL,
+    },
+    {
+        SCIM_SKK_CONFIG_COMPLETION_BACK_KEY,
+        SCIM_SKK_CONFIG_COMPLETION_BACK_KEY_DEFAULT,
+        N_("Completion Backward Keys:"),
+        N_("Select Completion Backward Keys"),
+        N_("The key events to turn back completion candidates of preedit string. "),
+        NULL,
+        NULL,
+    },
+    {
+        NULL,
+        "",
+        NULL,
+        NULL,
+        NULL,
+        NULL,
+        NULL,
+    },
+};
+
+static KeyboardConfigData __config_keyboards_conversion [] =
+{
     {
         SCIM_SKK_CONFIG_CONVERT_KEY,
         SCIM_SKK_CONFIG_CONVERT_KEY_DEFAULT,
@@ -230,15 +274,6 @@ static KeyboardConfigData __config_keyboards_common [] =
         NULL,
         NULL,
     },
-    {
-        SCIM_SKK_CONFIG_UPCASE_KEY,
-        SCIM_SKK_CONFIG_UPCASE_KEY_DEFAULT,
-        N_("Upcase Keys:"),
-        N_("Select Upcase Keys"),
-        N_("The key events to toggle case of characters in preedit string. "),
-        NULL,
-        NULL,
-    },
     {
         NULL,
         "",
@@ -347,8 +382,8 @@ static KeyboardConfigData __config_keyboards_caret [] =
         NULL,
     },
     {
-        SCIM_SKK_CONFIG_BACKWARD_KEY,
-        SCIM_SKK_CONFIG_BACKWARD_KEY_DEFAULT,
+        SCIM_SKK_CONFIG_END_KEY,
+        SCIM_SKK_CONFIG_END_KEY_DEFAULT,
         N_("End Keys:"),
         N_("Select End Keys"),
         N_("The key events to move the caret to the end of preedit. "),
@@ -383,8 +418,8 @@ static ComboConfigData selection_style[] =
 
 static ComboConfigData annot_position[] =
 {
-    {N_("aux window"), "auxwindow"},
-    {N_("inline"), "inline"},
+    {N_("Aux Window"), "AuxWindow"},
+    {N_("Inline"), "Inline"},
     {NULL, NULL},
 };
 
@@ -407,6 +442,7 @@ static ColorConfigData annot_bgcolor =
 static struct KeyboardConfigPage __key_conf_pages[] =
 {
     {N_("Common keys"),     __config_keyboards_common},
+    {N_("Conversion keys"), __config_keyboards_conversion},
     {N_("Mode keys"),       __config_keyboards_mode},
     {N_("Caret keys"),      __config_keyboards_caret},
 };
@@ -424,6 +460,8 @@ static void on_default_color_button_set       (GtkColorButton  *button,
                                                gpointer         user_data);
 static void on_default_file_selection_clicked (GtkButton       *button,
                                                gpointer         user_data);
+void on_default_dict_selection_clicked (GtkButton       *button,
+                                        gpointer         user_data);
 static void on_default_key_selection_clicked  (GtkButton       *button,
                                                gpointer         user_data);
 static void on_default_combo_changed          (GtkEditable     *editable,
@@ -529,7 +567,7 @@ create_options_page ()
     gtk_widget_show(widget);
     gtk_box_pack_start (GTK_BOX (vbox), widget, FALSE, FALSE, 4);
 
-    __widget_ignore_return = gtk_check_button_new_with_mnemonic(_("Ignore Return at the end of learning."));
+    __widget_ignore_return = gtk_check_button_new_with_mnemonic(_("Ignore Return at the commiting time."));
     gtk_widget_show(__widget_ignore_return);
     gtk_box_pack_start(GTK_BOX(vbox), __widget_ignore_return, FALSE, FALSE, 4);
 
@@ -599,28 +637,27 @@ create_options_page ()
 static GtkWidget *
 create_dictionary_page ()
 {
-    GtkWidget *vbox, *widget, *label;
+    GtkWidget *vbox, *widget, *label, *button;
 
     vbox = gtk_vbox_new (FALSE, 0);
     gtk_widget_show (vbox);
 
-    /* system dictionary */
+    /* system dictionaries */
     widget = gtk_hbox_new(FALSE, 0);
-    label  = gtk_label_new(_(__config_sysdict.title));
-    __config_sysdict.entry = gtk_entry_new();
-    __config_sysdict.button = gtk_button_new_with_label ("...");
-    gtk_widget_show (label);
-    gtk_widget_show (__config_sysdict.entry);
-    gtk_widget_show (__config_sysdict.button);
-    gtk_box_pack_start (GTK_BOX (widget), label, FALSE, FALSE, 4);
-    gtk_box_pack_start (GTK_BOX (widget), __config_sysdict.entry,
-                        TRUE, TRUE, 4);
-    gtk_box_pack_start (GTK_BOX (widget), __config_sysdict.button,
-                        FALSE, FALSE, 4);
-    gtk_widget_show (widget);
+    label = gtk_label_new(_("Configure Dictionaries"));
+    gtk_widget_show(label);
+    gtk_box_pack_start(GTK_BOX(widget), label, FALSE, FALSE, 4);
+    __widget_sysdicts = gtk_entry_new();
+    gtk_widget_show(__widget_sysdicts);
+    gtk_box_pack_start(GTK_BOX(widget), __widget_sysdicts, TRUE, TRUE, 4);
+    button = gtk_button_new_with_label(_("Configure Dictionaries"));
+    gtk_widget_show(button);
+    gtk_box_pack_start(GTK_BOX(widget), button, FALSE, FALSE, 4);
+    g_signal_connect ((gpointer) button, "clicked",
+                      G_CALLBACK (on_default_dict_selection_clicked),
+                      &__config_sysdicts);
+    gtk_widget_show(widget);
     gtk_box_pack_start (GTK_BOX (vbox), widget, FALSE, FALSE, 4);
-    gtk_label_set_mnemonic_widget (GTK_LABEL (label), __config_sysdict.entry);
-    gtk_label_set_mnemonic_widget (GTK_LABEL (label), __config_sysdict.button);
 
     /* user dictionary */
     widget            = gtk_hbox_new(FALSE, 0);
@@ -634,13 +671,6 @@ create_dictionary_page ()
     gtk_widget_show(widget);
     gtk_box_pack_start (GTK_BOX (vbox), widget, FALSE, FALSE, 4);
 
-    g_signal_connect ((gpointer) __config_sysdict.button, "clicked",
-                      G_CALLBACK (on_default_file_selection_clicked),
-                      &__config_sysdict);
-    g_signal_connect ((gpointer) __config_sysdict.entry, "changed",
-                      G_CALLBACK (on_default_editable_changed),
-                      &(__config_sysdict.data));
-
     g_signal_connect ((gpointer) __widget_userdict, "changed",
                       G_CALLBACK (on_default_editable_changed),
                       &__config_userdict);
@@ -794,6 +824,16 @@ setup_widget_value ()
                            annot_target, __config_annot_target);
     }
 
+    if (__widget_sysdicts && __config_sysdicts.size() > 0) {
+        std::vector<String>::iterator it = __config_sysdicts.begin();
+        String s = *it;
+        for (it++; it != __config_sysdicts.end(); it++) {
+            s.append(1, ',');
+            s.append(*it);
+        }
+        gtk_entry_set_text(GTK_ENTRY(__widget_sysdicts), s.data());
+    }
+    
     if (__widget_annot_view) {
         gtk_toggle_button_set_active (
             GTK_TOGGLE_BUTTON (__widget_annot_view),
@@ -850,9 +890,12 @@ static void
 load_config (const ConfigPointer &config)
 {
     if (!config.null ()) {
-        __config_sysdict.data =
+        std::vector<String> sysdicts_default;
+        scim_split_string_list (sysdicts_default,
+                                String (SCIM_SKK_CONFIG_SYSDICT_DEFAULT));
+        __config_sysdicts =
             config->read (String (SCIM_SKK_CONFIG_SYSDICT),
-                          __config_sysdict.data);
+                          sysdicts_default);
         __config_userdict =
             config->read (String (SCIM_SKK_CONFIG_USERDICT),
                           __config_userdict);
@@ -902,7 +945,7 @@ save_config (const ConfigPointer &config)
 {
     if (!config.null ()) {
         config->write (String (SCIM_SKK_CONFIG_SYSDICT),
-                       __config_sysdict.data);
+                       __config_sysdicts);
         config->write (String (SCIM_SKK_CONFIG_USERDICT),
                        __config_userdict);
         config->write (String (SCIM_SKK_CONFIG_CANDVEC_SIZE),
